"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var jsx_runtime_1 = require("react/jsx-runtime");
var react_1 = require("react");
var MIN_INDEX = 0;
var IS_OVERFLOW_ANCHOR_SUPPORTED = (function () {
    try {
        return window.CSS.supports('overflow-anchor: auto');
    }
    catch (error) {
        return false;
    }
})();
var PROP_NAME_FOR_Y_AXIS = {
    top: 'top',
    bottom: 'bottom',
    clientHeight: 'clientHeight',
    scrollTop: 'scrollTop',
    overflowY: 'overflowY',
    height: 'height',
    minHeight: 'minHeight',
    maxHeight: 'maxHeight'
};
var PROP_NAME_FOR_X_AXIS = {
    top: 'left',
    bottom: 'right',
    clientHeight: 'clientWidth',
    scrollTop: 'scrollLeft',
    overflowY: 'overflowX',
    height: 'width',
    minHeight: 'minWidth',
    maxHeight: 'maxWidth'
};
var getStyle = function (propName, size) {
    var _a;
    return (_a = {},
        _a[propName.minHeight] = size,
        _a[propName.height] = size,
        _a[propName.maxHeight] = size,
        _a.overflowAnchor = 'none',
        _a.pointerEvents = 'none',
        _a.userSelect = 'none',
        _a.padding = 0,
        _a.margin = 0,
        _a.border = 'none',
        _a);
};
var normalizeValue = function (min, value, max) { return Math.max(Math.min(value, max), min); };
var useRequestAnimationFrame = function (func) {
    var stepRef = (0, react_1.useRef)(func);
    stepRef.current = func;
    (0, react_1.useEffect)(function () {
        var frameId;
        var frame = function () {
            frameId = requestAnimationFrame(frame);
            stepRef.current();
        };
        frame();
        return function () {
            cancelAnimationFrame(frameId);
        };
    }, []);
};
var ViewportListInner = function (_a, ref) {
    var viewportRef = _a.viewportRef, _b = _a.items, items = _b === void 0 ? [] : _b, itemMinSize = _a.itemMinSize, _c = _a.margin, margin = _c === void 0 ? 0 : _c, _d = _a.overscan, overscan = _d === void 0 ? 1 : _d, _e = _a.axis, axis = _e === void 0 ? 'y' : _e, _f = _a.initialIndex, initialIndex = _f === void 0 ? 0 : _f, _g = _a.initialAlignToTop, initialAlignToTop = _g === void 0 ? true : _g, _h = _a.initialOffset, initialOffset = _h === void 0 ? 0 : _h, children = _a.children;
    var propName = axis === 'y' ? PROP_NAME_FOR_Y_AXIS : PROP_NAME_FOR_X_AXIS;
    var maxIndex = items.length - 1;
    var itemMinSizeWithMargin = itemMinSize + margin;
    var overscanSize = overscan * itemMinSizeWithMargin;
    var _j = (0, react_1.useState)(function () {
        var normalizedInitialIndex = normalizeValue(MIN_INDEX, initialIndex, maxIndex);
        return [normalizedInitialIndex, normalizedInitialIndex];
    }), _k = _j[0], startIndex = _k[0], endIndex = _k[1], setIndexes = _j[1];
    var normalizedStartIndex = normalizeValue(MIN_INDEX, startIndex, maxIndex);
    var normalizedEndIndex = normalizeValue(normalizedStartIndex, endIndex, maxIndex);
    var topRef = (0, react_1.useRef)(null);
    var bottomRef = (0, react_1.useRef)(null);
    var cacheRef = (0, react_1.useRef)([]);
    var scrollToIndexRef = (0, react_1.useRef)(startIndex ? { index: startIndex, alignToTop: initialAlignToTop, offset: initialOffset } : null);
    var scrollCompensationEndIndexRef = (0, react_1.useRef)(-1);
    var topStyle = (0, react_1.useMemo)(function () {
        return getStyle(propName, cacheRef.current
            .slice(MIN_INDEX, normalizedStartIndex)
            .reduce(function (sum, next) { return sum + next; }, normalizedStartIndex * itemMinSizeWithMargin));
    }, [normalizedStartIndex, itemMinSizeWithMargin, propName]);
    var bottomStyle = (0, react_1.useMemo)(function () {
        return getStyle(propName, cacheRef.current
            .slice(normalizedEndIndex + 1, maxIndex)
            .reduce(function (sum, next) { return sum + next; }, itemMinSizeWithMargin * (maxIndex - normalizedEndIndex)));
    }, [normalizedEndIndex, maxIndex, itemMinSizeWithMargin, propName]);
    var slicedItems = (0, react_1.useMemo)(function () {
        return items
            .slice(normalizedStartIndex, normalizedEndIndex + 1)
            .map(function (item, index) { return children(item, normalizedStartIndex + index, items); });
    }, [items, normalizedStartIndex, normalizedEndIndex, children]);
    useRequestAnimationFrame(function () {
        if (!viewportRef.current ||
            !topRef.current ||
            !bottomRef.current ||
            !(topRef.current.nextSibling instanceof Element) ||
            !(bottomRef.current.previousSibling instanceof Element)) {
            return;
        }
        var viewportRect = viewportRef.current.getBoundingClientRect();
        var topLimit = (viewportRect
            ? normalizeValue(0, viewportRect[propName.top], document.documentElement[propName.clientHeight])
            : 0) - overscanSize;
        var bottomLimit = (viewportRect
            ? normalizeValue(0, viewportRect[propName.bottom], document.documentElement[propName.clientHeight])
            : document.documentElement[propName.clientHeight]) + overscanSize;
        var topElementRect = topRef.current.nextSibling.getBoundingClientRect();
        var bottomElementRect = bottomRef.current.previousSibling.getBoundingClientRect();
        var maxItemsCountInViewPort = Math.ceil((bottomLimit - topLimit) / itemMinSizeWithMargin);
        var nextStartIndex = startIndex;
        var nextEndIndex = endIndex;
        var diff;
        var index;
        var element;
        if (scrollToIndexRef.current) {
            var targetIndex = normalizeValue(MIN_INDEX, scrollToIndexRef.current.index, maxIndex);
            if (targetIndex >= startIndex && targetIndex <= endIndex) {
                index = startIndex;
                element = topRef.current.nextSibling;
                while (element instanceof Element && element !== bottomRef.current) {
                    if (index === targetIndex) {
                        element.scrollIntoView(scrollToIndexRef.current.alignToTop);
                        if (scrollToIndexRef.current.offset) {
                            viewportRef.current[propName.scrollTop] += scrollToIndexRef.current.offset;
                        }
                        scrollToIndexRef.current = null;
                        break;
                    }
                    element = element.nextSibling;
                    ++index;
                }
                return;
            }
            nextStartIndex = targetIndex - maxItemsCountInViewPort;
            nextEndIndex = targetIndex + maxItemsCountInViewPort;
        }
        else if (topElementRect[propName.top] >= bottomLimit) {
            diff = topElementRect[propName.top] - bottomLimit;
            nextEndIndex = startIndex;
            while (diff >= 0 && nextEndIndex > MIN_INDEX) {
                diff -= (cacheRef.current[--nextEndIndex] || 0) + itemMinSizeWithMargin;
            }
            scrollCompensationEndIndexRef.current = startIndex;
            nextStartIndex = nextEndIndex - maxItemsCountInViewPort;
        }
        else if (bottomElementRect[propName.bottom] + margin <= topLimit) {
            diff = topLimit - bottomElementRect[propName.bottom] + margin;
            nextStartIndex = endIndex;
            while (diff >= 0 && nextStartIndex < maxIndex) {
                diff -= (cacheRef.current[++nextStartIndex] || 0) + itemMinSizeWithMargin;
            }
            nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
        }
        else {
            if (topElementRect[propName.bottom] + margin < topLimit) {
                ++nextStartIndex;
            }
            else if (topElementRect[propName.top] >= topLimit) {
                diff = topElementRect[propName.top] - topLimit;
                while (diff >= 0 && nextStartIndex > MIN_INDEX) {
                    diff -= (cacheRef.current[--nextStartIndex] || 0) + itemMinSizeWithMargin;
                }
                scrollCompensationEndIndexRef.current = startIndex;
            }
            if (bottomElementRect[propName.bottom] + margin <= bottomLimit) {
                diff = bottomLimit - bottomElementRect[propName.bottom] - margin;
                while (diff >= 0 && nextEndIndex < maxIndex) {
                    diff -= (cacheRef.current[++nextEndIndex] || 0) + itemMinSizeWithMargin;
                }
            }
            else if (bottomElementRect[propName.top] > bottomLimit) {
                --nextEndIndex;
            }
        }
        nextStartIndex = normalizeValue(MIN_INDEX, nextStartIndex, maxIndex);
        nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);
        if (nextStartIndex !== startIndex || nextEndIndex !== endIndex) {
            index = startIndex;
            element = topRef.current.nextSibling;
            while (element instanceof Element && index <= endIndex && element !== bottomRef.current) {
                cacheRef.current[index++] = element[propName.clientHeight] - itemMinSize;
                element = element.nextSibling;
            }
            setIndexes([nextStartIndex, nextEndIndex]);
        }
    });
    (0, react_1.useImperativeHandle)(ref, function () { return ({
        scrollToIndex: function (index, alignToTop, offset) {
            if (index === void 0) { index = -1; }
            if (alignToTop === void 0) { alignToTop = true; }
            if (offset === void 0) { offset = 0; }
            scrollToIndexRef.current = { index: index, alignToTop: alignToTop, offset: offset };
        }
    }); }, []);
    (0, react_1.useLayoutEffect)(function () {
        if (scrollCompensationEndIndexRef.current === -1) {
            return;
        }
        if (!viewportRef || !viewportRef.current || !topRef.current || IS_OVERFLOW_ANCHOR_SUPPORTED) {
            scrollCompensationEndIndexRef.current = -1;
            return;
        }
        var index = startIndex;
        var element = topRef.current.nextSibling;
        var sizeDiff = 0;
        while (element instanceof Element &&
            index < scrollCompensationEndIndexRef.current &&
            element !== bottomRef.current) {
            sizeDiff += element[propName.clientHeight] - (cacheRef.current[index] || 0) - itemMinSize;
            element = element.nextSibling;
            ++index;
        }
        if (sizeDiff) {
            viewportRef.current.style[propName.overflowY] = 'hidden';
            viewportRef.current[propName.scrollTop] += sizeDiff;
            viewportRef.current.style[propName.overflowY] = '';
        }
        scrollCompensationEndIndexRef.current = -1;
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [startIndex]);
    return ((0, jsx_runtime_1.jsxs)(react_1.Fragment, { children: [(0, jsx_runtime_1.jsx)("div", { ref: topRef, style: topStyle }, "ViewPortListTop"), slicedItems, (0, jsx_runtime_1.jsx)("div", { ref: bottomRef, style: bottomStyle }, "ViewPortListBottom")] }, void 0));
};
var ViewportList = (0, react_1.forwardRef)(ViewportListInner);
// eslint-disable-next-line import/no-default-export
exports.default = ViewportList;
